__all__ = ['Array', 'Dict', 'Document', 'FoamData', 'FoamItem', 'Keys', 'List', 'Location', 'Path', 'TupleSequence']


import pathlib as p
import typing as t

if t.TYPE_CHECKING:
    import numpy as _numpy

    from typing_extensions import Self


Dict = t.Dict[str, t.Any]
List = t.List[t.Any]
Document = FoamItem = t.Union[Dict, List]
FoamData = t.List[FoamItem]
Path = t.Union[str, p.Path]


class Array:
    '''
    TODO:
        - https://github.com/ramonhagenaars/nptyping?
    '''

    def __class_getitem__(cls, dimensions: 'Keys[int]') -> t.Union['_numpy.number', '_numpy.ndarray']:
        if not isinstance(dimensions, tuple):
            return cls.__class_getitem__((dimensions, ))
        assert all(isinstance(d, int) and d>=0 for d in dimensions)

        if dimensions == (0, ):
            return '_numpy.number'
        else:
            return '_numpy.ndarray'


class Keys:
    '''
    TODO:
        - https://stackoverflow.com/questions/47190218/proper-type-hint-for-getitem
    '''

    def __class_getitem__(cls, T: type) -> type:
        return t.Union[T, TupleSequence[T]]


class Location:
    '''Location 3D'''

    def __class_getitem__(cls, T: type) -> type:
        return t.Tuple[T, T, T]

    @classmethod
    def cast(cls, location: 'Self', T: type) -> 'Self':
        return tuple(map(T, location))


class TupleSequence:
    '''
    Reference:
        - https://github.com/python/mypy/issues/184
    '''

    def __class_getitem__(cls, T: type) -> type:
        return t.Tuple[T, ...]



__all__ = ['Parser']


import typing as t

from .case import Case
from .lark import Lark
from .static import Static
from .url import Url
from ..base.type import Path
from ..util.function import deprecated_classmethod

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ..base.core import Foam


class Parser:
    '''All parsers'''

    def __init__(self, case: Case, static: Static, url: Url) -> None:
        self.case = case
        self.static = static
        self.url = url

    @classmethod
    def fromFoam(cls, foam: 'Foam') -> 'Self':
        return cls(
            Case.default(),
            Static.fromFoam(foam),
            Url.fromFoam(foam),
        )

    @classmethod
    def toFoam(cls, path: Path, **kwargs: t.Any) -> 'Foam':
        from ..base.core import Foam

        data = Lark.fromPath(path, **kwargs).to_foam_data()
        data[0]['version'] = Foam.__version__.to_string()
        return Foam(data, path)

    from_foam = deprecated_classmethod(fromFoam)
    to_foam = deprecated_classmethod(toFoam)



__all__ = ['Lark']


import os
import pathlib as p
import re
import typing as t

from ..base.lib import lark
from ..base.type import Dict, FoamData, List, Path
from ..compat.functools import cached_property
from ..util.function import deprecated_classmethod, grammar
from ..util.implementation import Singleton

if t.TYPE_CHECKING:
    from typing_extensions import Self

    import lark as _lark


class Lark(Singleton):
    '''Lark is a parsing toolkit for Python

    TODO:
        - unit-test (Lark, Foam.fromOpenFoam)
    '''

    order = ['meta', 'foam', 'static', 'other']

    def __init__(self, path: Path, embed: bool = True) -> None:
        self._root = p.Path(path)
        self._embed = embed
        # foam, static
        self._foam = {}
        self._static = []

    @classmethod
    def fromPath(cls, path: Path, **kwargs: t.Any) -> 'Self':
        return cls.new(path, **kwargs)

    @property
    def meta(self) -> Dict:
        return {'openfoam': [self._openfoam()], 'version': '0.0.0', 'order': self.order}

    @property
    def foam(self) -> Dict:
        return self._foam

    @property
    def static(self) -> List:
        return self._static

    @property
    def other(self) -> Dict:
        return {'pipeline': []}

    @cached_property
    def lark(self) -> '_lark.Lark':
        kwargs = {
            'debug': False,
            'cache': False,  # cache only works with parser='lalr' for now
            'parser': 'earley', # or 'lalr'
            'lexer': 'dynamic',  # Parser 'lalr' does not support lexer 'dynamic', expected one of ('basic', 'contextual')
            'transformer': None,
            'start': 'start',
        }
        return lark.Lark(grammar(), **kwargs)

    def parse(self) -> None:
        for path in self._root.rglob('*'):
            if path.is_file():
                # TODO: foam part
                self._static.append(self._static_item(path))

    def parse_once(self) -> None:
        '''#pragma once'''
        if not self.parsed():
            self.parse()

    def parsed(self) -> bool:
        return bool(self._foam) or bool(self._static)

    def to_foam_data(self) -> FoamData:
        self.parse_once()
        return [getattr(self, o) for o in self.order]

    def _openfoam(self) -> str:
        '''OpenFOAM Version'''
        pattern = re.compile(r'^(OpenFOAM-)([\d.x]+)$')
        for part in self._root.parts:
            match = pattern.match(part)
            if match is not None:
                return match.groups()[-1]
        return os.environ['WM_PROJECT_VERSION']

    def _static_item(self, path: p.Path) -> Dict:
        name = path.relative_to(self._root).as_posix()
        permission = oct(path.stat().st_mode)[-3:]
        if self._embed:
            data = path.read_bytes()
            try:
                data = data.decode()
            except UnicodeDecodeError:
                type = ['embed', 'binary']
            else:
                type = ['embed', 'text']
        else:
            data = path.as_posix()
            type = ['path', 'raw']
        return {'name': name, 'type': type, 'permission': permission, 'data': data}

    from_path = deprecated_classmethod(fromPath)



__all__ = ['copytree']


import shutil


def copytree(
    src, dst,
    symlinks=False, ignore=None, copy_function=shutil.copy2,
    ignore_dangling_symlinks=False, dirs_exist_ok=False,
):
    try:
        return shutil.copytree(
            src, dst,
            symlinks=symlinks, ignore=ignore, copy_function=copy_function,
            ignore_dangling_symlinks=ignore_dangling_symlinks,
        )
    except FileExistsError as e:
        if not dirs_exist_ok:
            raise e



__all__ = ['cached_property', 'singledispatchmethod']


import functools as f
import sys


version = (sys.version_info.major, sys.version_info.minor)
if version == (3, 7):
    from .v37 import cached_property, singledispatchmethod
elif version == (3, 8):
    cached_property = f.cached_property
    singledispatchmethod = f.singledispatchmethod
elif version == (3, 9):
    cached_property = f.cached_property
    singledispatchmethod = f.singledispatchmethod
elif version == (3, 10):
    cached_property = f.cached_property
    singledispatchmethod = f.singledispatchmethod
else:
    raise Exception('Current version not supported')



__all__ = ['Singleton']


import typing as t

from ..compat.typing import Protocol

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Singleton(Protocol):
    '''Singleton'''

    _instances = {}

    @classmethod
    def default(cls) -> 'Self':
        return cls.new()

    @classmethod
    def new(cls, *args: t.Any, **kwargs: t.Any) -> 'Self':
        key = cls._hash(*args, **kwargs)
        if key not in cls._instances:
            cls._instances[key] = cls(*args, **kwargs)
        return cls._instances[key]

    @classmethod
    def _hash(cls, *args: t.Any, **kwargs: t.Any) -> str:
        return repr(args) + repr(kwargs)



__all__ = ['Envelope', 'SMTP']


import email.message
import email.utils
import mimetypes
import pathlib as p
import smtplib
import time
import typing as t

from ...base.type import Path

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Envelope:
    '''Envelope

    Reference:
        - https://github.com/tomekwojcik/envelopes
    '''

    _smtp = None

    def __init__(self, *to_addresses: str) -> None:
        self._header = {'To': ', '.join(to_addresses)}
        self._applies = [lambda msg: None]

    @classmethod
    def auto(cls, smtp: 'SMTP') -> 'Self':  # type?
        '''Auto-delivered envelope'''
        cls._smtp = smtp
        return cls

    @classmethod
    def to(cls, *addresses: str) -> 'Self':
        return cls(*addresses)

    def add_attachment(self, path: Path, type: t.Optional[str] = None, **kwargs: t.Any) -> 'Self':
        '''
        Reference:
            - https://docs.python.org/zh-cn/3/library/email.examples.html
        '''
        path = p.Path(path)
        if type is None:
            type, encoding = mimetypes.guess_type(path.as_posix(), strict=True)
            if type is None or encoding is not None:
                type = 'application/octet-stream'
        maintype, subtype = type.split('/', maxsplit=1)
        content = path.read_bytes()
        kwargs = {'maintype': maintype, 'subtype': subtype, 'filename': path.name, **kwargs}
        self._applies.append(lambda msg: msg.add_attachment(content, **kwargs))
        return self

    def set(self, **kwargs: str) -> 'Self':
        for key, value in kwargs.items():
            getattr(self, f'set_{key}')(value)
        return self

    def set_date(self, value: t.Optional[float]) -> 'Self':
        self._header['Date'] = email.utils.formatdate(value, localtime=True, usegmt=False)

    def set_content(self, value: str, html: bool = False) -> 'Self':
        self._applies[0] = lambda msg: msg.set_content(value, subtype='html' if html else 'plain')
        return self

    def set_content_by_path(self, value: Path, html: bool = False) -> 'Self':
        return self.set_content(p.Path(value).read_text(), html)

    def set_subject(self, value: str) -> 'Self':
        self._header['Subject'] = value
        return self

    def to_message(self) -> email.message.EmailMessage:
        msg = email.message.EmailMessage()
        # _header
        for key, value in self._header.items():
            msg[key] = value
        # _applies
        for apply in self._applies:
            apply(msg)
        return msg

    def send_by(self, *smtps: 'SMTP') -> 'Self':
        for smtp in smtps:
            smtp.send(self)
        return self

    def send(self) -> None:
        assert self._smtp is not None, 'Please do not use this class directly'

        self.send_by(self._smtp)


class SMTP:
    '''SMTP wrapper

    Example:
        >>> with SMTP.aio('163', username, password, ssl=False) as smtp:
        ...     smtp.envelope \\
        ...         .to('liangiydon@gmail.com') \\
        ...         .set_subject('SMTP Test') \\
        ...         .set_content('Here is the <a href="http://www.python.org">link</a> you wanted.', html=True) \\
        ...         .add_attachment(__file__) \\
        ...         .send()
    '''

    def __init__(self, domain: str, host: str, port: int, ssl: bool = True) -> None:
        self._domain = domain
        self._smtp = (smtplib.SMTP_SSL if ssl else smtplib.SMTP)(host, port)
        self._username = None

    def __enter__(self) -> 'Self':
        return self

    def __exit__(self, type, value, traceback) -> None:
        self._smtp.__exit__(type, value, traceback)

    def __class_getitem__(cls, key: str) -> 'Self':
        # TODO: look forward to adding more
        return {
            '163': lambda ssl: cls('163.com', 'smtp.163.com', 25, ssl),  # mail.163.com
            'qq': lambda ssl: cls('qq.com', 'smtp.qq.com', 587, ssl),  # mail.qq.com
        }[key]

    @classmethod
    def aio(cls, mail: str, username: str, password: str, ssl: bool = True) -> 'Self':
        '''All-in-one'''
        return cls[mail](ssl).login(username, password)

    @property
    def sender(self) -> str:
        assert self._username is not None, 'Please login first'

        return f'{self._username}@{self._domain}'

    @property
    def envelope(self) -> 'Envelope':
        return Envelope.auto(self)

    def login(self, username: str, password: str) -> 'Self':
        self._username = username
        self._smtp.login(username, password)
        return self

    def quit(self) -> None:
        # NOTE: lifetime of this instance ends after quit is called, so it will not return self
        try:
            self._smtp.quit()
        except smtplib.SMTPServerDisconnected:
            pass

    def send(self, *envelopes: 'Envelope') -> 'Self':
        for envelope in envelopes:
            msg = envelope.to_message()
            msg['From'] = self.sender
            self._smtp.send_message(msg)
        return self

    def wait(self, seconds: float) -> 'Self':
        time.sleep(seconds)
        return self



__all__ = ['copytree']


import shutil
import sys


version = (sys.version_info.major, sys.version_info.minor)
if version == (3, 7):
    from .v37 import copytree
elif version == (3, 8):
    copytree = shutil.copytree
elif version == (3, 9):
    copytree = shutil.copytree
elif version == (3, 10):
    copytree = shutil.copytree
else:
    raise Exception('Current version not supported')



__all__ = ['END', 'End', 'end']


class end:
    '''End of indented code block

    Example:
        ```python
        class func:
            @staticmethod
            def reciprocal(x: int) -> float:
                return 1.0 / x
            End.Def
        End.Class

        for ith in range(10):
            if ith < 7:
                try:
                    print(func.reciprocal(ith))
                except ZeroDivisionError:
                    print('NaN')
                End.Try
            End.If
        End.For
        ```

    Reference:
        - https://docs.python.org/3/reference/compound_stmts.html

    Statement:
        - [x] if_stmt
        - [x] while_stmt
        - [x] for_stmt
        - [x] try_stmt
        - [x] with_stmt
        - [x] match_stmt
        - [x] funcdef
        - [x] classdef
        - [x] async_with_stmt
        - [x] async_for_stmt
        - [x] async_funcdef
    '''

    IF = If = if_ = None
    WHILE = While = while_ = None
    FOR = For = for_ = None
    TRY = Try = try_ = None
    WITH = With = with_ = None
    MATCH = Match = match_ = None
    CASE = Case = case_ = None  # placeholder only
    DEF = Def = def_ = None
    CLASS = Class = class_ = None
    ASYNC_WITH = AsyncWith = async_with = None
    ASYNC_FOR = AsyncFor = async_for = None
    ASYNC_DEF = AsyncDef = async_def = None


END = End = end



__all__ = ['Match', 'classproperty', 'message', 'suppress']


import contextlib
import functools as f
import io
import typing as t

from ..base.lib import classproperty
from ..base.type import Keys, TupleSequence

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Match:
    '''Use decorator to simulate match syntax

    Example:
        >>> match = Match.default()

        >>> @match.register()
        ... def _():
        ...     return 0

        >>> @match.register('x')
        ... def _():
        ...     return 1

        >>> @match.register('x', 'y')
        ... def _():
        ...     return 2

        >>> match[()]()
        0
        >>> match['x']()
        1
        >>> match['x', 'y']()
        2
    '''

    def __init__(self) -> None:
        self._methods: t.Dict[TupleSequence[str], t.Callable] = {}

    def __getitem__(self, keys: Keys[str]) -> t.Callable:
        if not isinstance(keys, tuple):
            return self.__getitem__((keys, ))
        return self._methods[keys]

    @classmethod
    def default(cls) -> 'Self':
        return cls()

    def get(self, *keys: str, default: t.Optional[t.Callable] = None) -> t.Optional[t.Callable]:
        return self._methods.get(keys, default)

    def register(self, *types: str) -> t.Callable:
        '''Register methods for different types'''

        def decorate(func: t.Callable) -> t.Callable:
            self._methods[types] = func

            @f.wraps(func)
            def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
                return func(*args, **kwargs)

            return wrapper

        return decorate


class suppress:
    '''Suppress stderr or stdout

    Reference:
        - https://docs.python.org/3/library/unittest.html
        - https://stackoverflow.com/questions/9718950/do-i-have-to-do-stringio-close
    '''

    class _base:
        '''Base'''

        __redirect__: t.Callable

        _previous = ''

        @classmethod
        @contextlib.contextmanager
        def contextWithoutPrevious(cls) -> t.Iterator[io.StringIO]:
            with io.StringIO() as target, cls.__redirect__(target):
                try:
                    yield target
                finally:
                    pass

        @classmethod
        @contextlib.contextmanager
        def contextWithPrevious(cls) -> t.Iterator[io.StringIO]:
            with cls.contextWithoutPrevious() as target:
                try:
                    yield target
                finally:
                    cls._previous = target.getvalue()

        @classmethod
        def decoratorWithoutPrevious(cls, func: t.Callable) -> t.Callable:

            @f.wraps(func)
            def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
                with cls.contextWithoutPrevious():
                    return func(*args, **kwargs)

            return wrapper

        @classmethod
        def decoratorWithPrevious(cls, func: t.Callable) -> t.Callable:

            @f.wraps(func)
            def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
                with cls.contextWithPrevious():
                    return func(*args, **kwargs)

            return wrapper

        @classproperty
        def previous(cls) -> str:
            return cls._previous

        context_without_previous = contextWithoutPrevious
        context_with_previous = contextWithPrevious
        decorator_without_previous = decoratorWithoutPrevious
        decorator_with_previous = decoratorWithPrevious

        context = contextWithoutPrevious
        decorator = decoratorWithPrevious

    class stderr(_base):
        '''Std Err'''

        __redirect__ = contextlib.redirect_stderr

    class stdout(stderr):
        '''Std Out'''

        __redirect__ = contextlib.redirect_stdout


def message(msg: str = '') -> t.Callable:
    
    def decorate(func: t.Callable) -> t.Callable:

        @f.wraps(func)
        def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            print(msg)
            return func(*args, **kwargs)

        return wrapper

    return decorate



__all__ = ['cnv', 'run']


import pathlib as p
import subprocess

import click

from foam import Foam
from foam.base.lib import tqdm
from foam.base.type import TupleSequence


class DEFAULT:
    '''Default arguments'''

    DIRECTORY = 'test'
    OPENFOAM = '7'


def _cnv(src: p.Path, dst: p.Path, version: str, exist_ok: bool) -> bool:
    try:
        foam = Foam.fromPath(src)
    except Exception as e:
        click.echo(f'{e!r}: {src}')
        return False
    else:
        if version in set(map(str, foam.meta.get('openfoam', []))):
            path = dst / '-'.join(src.parts)
            if exist_ok or not path.exists():
                foam.save(path)
        return True


def _run(directory: p.Path) -> bool:
    cp = subprocess.run('./Allrun', capture_output=True, cwd=directory, shell=True)
    return cp.returncode == 0


@click.group()
@click.version_option(version=Foam.__version__.to_string(), prog_name=Foam.__name__)
def cli() -> None:
    pass


@cli.command(help=Foam.__doc__)
@click.argument('paths', nargs=-1)  # YAML format files or directories
@click.option('-d', '--directory', default=DEFAULT.DIRECTORY, help='Destination directory')
@click.option('-v', '--version', default=DEFAULT.OPENFOAM, help='OpenFOAM version')
@click.option('-o', '--exist-ok', is_flag=True, help='If `exist_ok` then do not overwrite')
def cnv(paths: TupleSequence[str], directory: str = DEFAULT.DIRECTORY, version: str = DEFAULT.OPENFOAM, exist_ok: bool = True) -> None:
    dst = p.Path(directory, version)
    for path in paths:
        src = p.Path(path)
        if src.is_dir():
            for path in src.rglob('*'):
                if path.is_file() and path.suffix in {'.yaml', '.yml'}:
                    assert _cnv(path, dst, version, exist_ok), path
        elif src.is_file():
            assert _cnv(src, dst, version, exist_ok), src
        else:
            raise Exception(f'Path "{src.as_posix()}" does not exist, or is neither a file nor a directory')


@cli.command()
@click.option('-d', '--directory', default=DEFAULT.DIRECTORY, help='Directory containing YAML format files')
@click.option('-v', '--version', default=DEFAULT.OPENFOAM, help='OpenFOAM version')
def run(
    directory: str = DEFAULT.DIRECTORY,
    version: str = DEFAULT.OPENFOAM,
) -> None:
    '''For batch testing whether the converted files are operational'''
    pbar = tqdm.tqdm if tqdm.is_available() else lambda x: x
    dst = p.Path(directory, version)
    for directory in pbar(dst.iterdir()):
        if directory.is_dir():
            assert _run(directory), directory


if __name__ == '__main__':
    cli()



__all__ = ['Figure']


import pathlib as p
import typing as t

from ..function import deprecated_classmethod
from ...base.lib import matplotlib
from ...base.type import Path

if t.TYPE_CHECKING:
    import matplotlib.axes as _axes
    import matplotlib.figure as _figure

    from typing_extensions import Self


class Figure:
    '''Matplotlib simple wrapper

    Example:
        >>> Figure.setRcParams(
        ...     ('axes.unicode_minus', False),
        ...     ('font.sans-serif', ['FandolHei']),
        ... )
        >>> Figure.new(figsize=(8, 6)) \\
        ...     .plot([1, 2, 3], [3, 2, 1], label='a') \\
        ...     .plot([1, 2, 3], [2, 1, 3], label='b') \\
        ...     .set_ax(xlabel='x label', ylabel='y label', title='title') \\
        ...     .grid() \\
        ...     .legend() \\
        ...     .save('demo.png')
    '''

    DEFAULT = {
        'legend': {
            'bbox_to_anchor': (1.01, 1), 'loc': 'upper left',
            'borderaxespad': 0, 'ncol': 1,
        },
        'save': {'bbox_inches': 'tight', 'transparent': False},
    }

    _fig: '_figure.Figure'
    _ax: '_axes.SubplotBase'

    def __init__(self, **kwargs: t.Any) -> None:
        self._fig, self._ax = matplotlib.pyplot.subplots(1, 1, **kwargs)

    @classmethod
    def new(cls, **kwargs: t.Any) -> 'Self':
        return cls(**kwargs)

    @classmethod
    def setRcParams(cls, *items: t.Tuple[str, t.Any]) -> 'Self':
        # Type annotation of return value should actually be type
        plt = matplotlib.pyplot
        for key, value in items:
            plt.rcParams[key] = value
        return cls

    def grid(self, *args: t.Any, **kwargs: t.Any) -> 'Self':
        self._ax.grid(*args, **kwargs)
        return self

    def legend(self, *args: t.Any, **kwargs: t.Any) -> 'Self':
        kwargs = {**self.DEFAULT['legend'], **kwargs}
        self._ax.legend(*args, **kwargs)
        return self

    def plot(self, *args: t.Any, **kwargs: t.Any) -> 'Self':
        self._ax.plot(*args, **kwargs)
        return self

    def save(self, path: Path, **kwargs: t.Any) -> 'Self':
        kwargs = {**self.DEFAULT['save'], **kwargs}
        self._fig.savefig(p.Path(path).as_posix(), **kwargs)
        return self

    def set_ax(self, *args: t.Any, **kwargs: t.Any) -> 'Self':
        self._ax.set(*args, **kwargs)
        return self

    set_rc_params = deprecated_classmethod(setRcParams)



__all__ = ['Protocol']


import sys
import typing as t


version = (sys.version_info.major, sys.version_info.minor)
if version == (3, 7):
    from .v37 import Protocol
elif version == (3, 8):
    Protocol = t.Protocol
elif version == (3, 9):
    Protocol = t.Protocol
elif version == (3, 10):
    Protocol = t.Protocol
else:
    raise Exception('Current version not supported')



__all_ = ['Origin', 'DryRun']


import subprocess
import typing as t

if t.TYPE_CHECKING:
    from typing_extensions import Self


Origin = subprocess.Popen


class DryRun:
    '''Dry run'''

    args = None
    stdout = b''
    returncode = 0

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        print(f'Popen({self._repr(*args, **kwargs)})')

    def __enter__(self) -> 'Self':
        return self

    def __exit__(self, type, value, traceback) -> None:
        pass

    def communicate(self, *args: t.Any, **kwargs: t.Any) -> None:
        print(f'Popen.communicate({self._repr(*args, **kwargs)})')
        return (b'', b'')

    def poll(self) -> int:
        return self.returncode

    def _repr(self, *args: t.Any, **kwargs: t.Any) -> str:
        # caller_func_name = sys._getframe(1).f_code.co_name
        return f'{self._repr_args(*args)}, {self._repr_kwargs(**kwargs)}'

    def _repr_args(self, *args: t.Any) -> str:
        return ', '.join(repr(arg) for arg in args)

    def _repr_kwargs(self, **kwargs: t.Any) -> str:
        return ', '.join(f'{k!s}={v!r}' for k, v in kwargs.items())



__all__ = ['Command']


import pathlib as p
import shlex
import shutil
import subprocess as s
import typing as t
import warnings as w

from .adapter import Default, Apps
from ...base.type import Dict
from ...compat.functools import cached_property
from ...util.function import deprecated_classmethod

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ...base.core import Foam


class Command:
    '''OpenFOAM command wrapper'''

    def __init__(self, foam: 'Foam') -> None:
        self._foam = foam

    @classmethod
    def fromFoam(cls, foam: 'Foam') -> 'Self':
        foam.destination  # assert dest is not None
        return cls.fromFoamWithoutAsserting(foam)

    @classmethod
    def fromFoamWithoutAsserting(cls, foam: 'Foam') -> 'Self':
        return cls(foam)

    @property
    def times(self) -> t.List[float]:
        '''Time directories'''
        times = []
        for path in self._foam.destination.iterdir():
            try:
                time = float(path.name)
            except ValueError:
                continue
            times.append(time)
        return sorted(times)

    @property
    def logs(self) -> t.Set[p.Path]:
        '''Log files'''
        logs = set()
        for path in self._foam.destination.iterdir():
            if path.stem == 'log':
                logs.add(path)
        return logs

    @cached_property
    def macros(self) -> t.Dict[str, str]:
        '''Macros that can be used in the pipeline field'''
        macros = {
            '__app__': self._foam.application,
            '__procs__': str(self._foam.number_of_processors),
        }
        try:
            macros['__pwd__'] = self._foam.destination.absolute().as_posix()
        except AssertionError:
            pass
        return macros

    def all_run(
        self,
        overwrite: bool = False, exception: bool = False,
        parallel: bool = True, unsafe: bool = True,
    ) -> t.List[int]:
        '''Inspired by  `Allrun`'''
        if not self._foam.pipeline:
            assert (self._foam.destination/'Allrun').exists()

            return [self.raw('./Allrun').returncode]
        else:
            return self.run(self._foam.pipeline, overwrite=overwrite, exception=exception, parallel=parallel, unsafe=unsafe)

    def all_clean(self) -> None:
        '''Inspired by `Allclean`'''
        # TODO: https://github.com/OpenFOAM/OpenFOAM-7/blob/master/bin/tools/CleanFunctions
        shutil.rmtree(self._foam.destination)
        self._foam.save(self._foam.destination)

    def run(
        self,
        commands: t.List[t.Union[str, Dict]],
        suffix: str = '', overwrite: bool = False, exception: bool = True,
        parallel: bool = True, unsafe: bool = False,
    ) -> t.List[int]:
        '''Inspired by `runApplication` and `runParallel`

        Reference:
            - https://github.com/OpenFOAM/OpenFOAM-7/blob/master/bin/tools/RunFunctions
        '''
        codes: t.List[int] = [-1] * len(commands)
        for ith, command in enumerate(commands):
            option = self._command(command, suffix=suffix, overwrite=overwrite, exception=exception, parallel=parallel)
            raws = self._split(option['command'], False)
            args = self._split(option['command'], option['parallel'] and self._foam.number_of_processors>1)
            path = self._foam.destination / f'log.{raws[0].replace("./", "")}{option["suffix"]}'
            # Option: overwrite, exception
            if not option['overwrite'] and path.exists():
                message = f'{raws[0]} already run on {path.parent.absolute()}: remove log file "{path.name}" to re-run'
                if option['exception']:
                    raise Exception(message)
                else:
                    w.warn(message)
                    continue
            # TODO: verbose?
            print(f'Running {raws[0]} on {path.parent.absolute()} using {self._foam.number_of_processors} processes if in parallel')
            # TODO: rewritten as parenthesized context managers when updated to 3.10
            App = Apps.get(raws[0], Default)
            with self._popen(args, unsafe) as proc, open(path, 'wb') as file, App(self._foam) as app:
                for line in proc.stdout:
                    file.write(line)
                    app.step(line)
            codes[ith] = proc.returncode
        return codes

    def raw(self, command: str, output: bool = True) -> s.CompletedProcess:
        '''Execute raw command in case directory'''
        args = shlex.split(command)
        return s.run(args, cwd=self._foam._dest, capture_output=output)  # Deliberate use of the _dest

    def which(self, command: str) -> t.Optional[str]:
        stdout = self.raw(f'which {command}', output=True).stdout.decode().strip()
        return stdout or None

    def _replace(self, command: str) -> str:
        for old, new in self.macros.items():
            command = command.replace(old, new)
        return command

    def _split(self, command: str, parallel: bool, replace: bool = True) -> t.List[str]:
        if parallel and '__app__' in command:
            command = f'mpirun -np __procs__ {command} -parallel'
        if replace:
            command = self._replace(command)
        return shlex.split(command)

    def _command(self, command: t.Union[str, Dict], **kwargs: t.Any) -> Dict:
        if isinstance(command, str):
            return {'command': command, **kwargs}
        elif isinstance(command, dict):
            kwargs.update(command)
            return kwargs
        else:
            raise Exception('`command` does not currently support variables other than `str`, `dict`')

    def _popen(self, args: t.List[str], unsafe: bool) -> s.Popen:
        cmd = ' '.join(args) if unsafe else args
        return s.Popen(cmd, cwd=self._foam.destination, shell=unsafe, stdout=s.PIPE)

    from_foam = deprecated_classmethod(fromFoam)
    from_foam_without_asserting = deprecated_classmethod(fromFoamWithoutAsserting)



__all__ = ['classproperty', 'lark', 'matplotlib', 'numpy', 'py7zr', 'tomlkit', 'tqdm', 'vtkmodules', 'yaml']


import typing as t

if t.TYPE_CHECKING:
    import lark as _lark
    import matplotlib as _matplotlib
    import matplotlib.figure as _figure
    import matplotlib.pyplot as _pyplot
    import numpy as _numpy
    import py7zr as _py7zr
    import tomlkit as _tomlkit
    import tqdm as _tqdm
    import vtkmodules as _vtkmodules
    import yaml as _yaml

    from typing_extensions import Self


class classproperty:
    '''Decorator that converts a method with a single cls argument into a property that can be accessed directly from the class.

    Reference:
        - https://github.com/django/django

    TODO:
        - place it in util::decorator will cause circular import problem
    '''

    def __init__(self, method: t.Optional[t.Callable] = None) -> None:
        self.fget = method

    def __get__(self, instance: t.Any, cls: t.Optional[type] = None) -> t.Any:
        return self.fget(cls)

    def getter(self, method: t.Callable) -> 'Self':
        self.fget = method
        return self


class lark:
    '''pip install ifoam[lark]'''

    @classmethod
    def Lark(cls, *args: t.Any, **kwargs: t.Any) -> '_lark.Lark':
        return cls._().Lark(*args, **kwargs)

    @classmethod
    def _(cls) -> '_lark':
        try:
            import lark
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return lark


class matplotlib:
    '''pip install ifoam[mpl]'''

    @classproperty
    def figure(cls) -> '_figure':
        return cls._().figure

    @classproperty
    def pyplot(cls) -> '_pyplot':
        return cls._().pyplot

    @classmethod
    def _(cls) -> '_matplotlib':
        try:
            import matplotlib
            import matplotlib.figure
            import matplotlib.pyplot
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return matplotlib


class numpy:
    '''pip install ifoam[vtk]'''

    @classmethod
    def argmin(cls, *args: t.Any, **kwargs: t.Any) -> '_numpy.ndarray':
        return cls._().argmin(*args, **kwargs)

    @classmethod
    def loadtxt(cls, *args: t.Any, **kwargs: t.Any) -> '_numpy.ndarray':
        return cls._().loadtxt(*args, **kwargs)

    @classmethod
    def square(cls, *args: t.Any, **kwargs: t.Any) -> '_numpy.ndarray':
        return cls._().square(*args, **kwargs)

    @classmethod
    def _(cls) -> '_numpy':
        try:
            import numpy
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return numpy


class py7zr:
    '''pip install ifoam[7z]'''

    @classmethod
    def SevenZipFile(cls, *args: t.Any, **kwargs: t.Any) -> '_py7zr.SevenZipFile':
        return cls._().SevenZipFile(*args, **kwargs)

    @classmethod
    def _(cls) -> '_py7zr':
        try:
            import py7zr
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return py7zr


class tomlkit:
    '''pip install ifoam[toml]'''

    @classmethod
    def dump(cls, *args: t.Any, **kwargs: t.Any) -> None:
        cls._().dump(*args, **kwargs)

    @classmethod
    def dumps(cls, *args: t.Any, **kwargs: t.Any) -> str:
        return cls._().dumps(*args, **kwargs)

    @classmethod
    def load(cls, *args: t.Any, **kwargs: t.Any) -> '_tomlkit.toml_document.TOMLDocument':
        return cls._().load(*args, **kwargs)

    @classmethod
    def loads(cls, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return cls._().loads(*args, **kwargs)

    @classmethod
    def _(cls) -> '_tomlkit':
        try:
            import tomlkit
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return tomlkit


class tqdm:
    '''pip install ifoam[tqdm]'''

    @classmethod
    def is_available(cls) -> bool:
        try:
            cls._()
        except Exception:
            return False
        else:
            return True

    @classmethod
    def is_not_available(cls) -> bool:
        return not cls.is_available()

    @classmethod
    def tqdm(cls, *args: t.Any, **kwargs: t.Any) -> '_tqdm.std.tqdm':
        return cls._().tqdm(*args, **kwargs)

    @classmethod
    def _(cls) -> '_tqdm':
        try:
            import tqdm
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return tqdm


class vtkmodules:
    '''pip install ifoam[vtk]'''

    @classmethod
    def vtkGenericDataObjectReader(cls, *args: t.Any, **kwargs: t.Any) -> '_vtkmodules.vtkIOLegacy.vtkGenericDataObjectReader':
        return cls._().vtkIOLegacy.vtkGenericDataObjectReader()

    @classmethod
    def vtk_to_numpy(cls, *args: t.Any, **kwargs: t.Any) -> '_numpy.ndarray':
        try:
            from vtkmodules.util.numpy_support import vtk_to_numpy
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return vtk_to_numpy(*args, **kwargs)

    @classmethod
    def _(cls) -> '_vtkmodules':
        try:
            import vtkmodules.all
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return vtkmodules


class yaml:
    '''pip install ifoam'''

    @classmethod
    def dump(cls, *args: t.Any, **kwargs: t.Any) -> str:
        kwargs = {'Dumper': cls._dumper(), **kwargs}
        return cls._().dump(*args, **kwargs)

    @classmethod
    def dump_all(cls, *args: t.Any, **kwargs: t.Any) -> str:
        kwargs = {'Dumper': cls._dumper(), **kwargs}
        return cls._().dump_all(*args, **kwargs)

    @classmethod
    def load(cls, *args: t.Any, **kwargs: t.Any) -> t.Any:
        kwargs = {'Loader': cls._loader(), **kwargs}
        return cls._().load(*args, **kwargs)

    @classmethod
    def load_all(cls, *args: t.Any, **kwargs: t.Any) -> t.Iterator[t.Any]:
        kwargs = {'Loader': cls._loader(), **kwargs}
        return cls._().load_all(*args, **kwargs)

    @classmethod
    def _(cls) -> '_yaml':
        try:
            import yaml
        except Exception as e:
            raise e.__class__(cls.__doc__)
        else:
            return yaml

    @classmethod
    def _dumper(cls) -> t.Union['_yaml.SafeDumper', '_yaml.CSafeDumper']:
        yaml = cls._()
        if yaml.__with_libyaml__:
            return yaml.CSafeDumper
        else:
            return yaml.SafeDumper

    @classmethod
    def _loader(cls) -> t.Union['_yaml.SafeLoader', '_yaml.CSafeLoader']:
        yaml = cls._()
        if yaml.__with_libyaml__:
            return yaml.CSafeLoader
        else:
            return yaml.SafeLoader



import importlib as _importlib

from .namespace import *


__import__ = lambda name: _importlib.import_module(name, __name__)
__all__ = __import__('.namespace').__all__
__doc__ = Foam.__doc__
__license__ = __import__('.util.function').license(full_text=False)
__version__ = Foam.__version__



__all__ = ['Conversion']


import json
import pathlib as p
import pickle
import typing as t

from ..function import deprecated_classmethod
from ...base.lib import tomlkit, yaml
from ...base.type import Document, Path

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Conversion:
    '''Conversion between object and bytes/string

    Example:
        >>> data = {'a': 1, 'b': [2, 3], 'c': {'4': 5}}
        >>> print(Conversion.fromDocument(data).to_yaml())
        a: 1
        b:
        - 2
        - 3
        c:
        '4': 5
    '''

    _alias = {'pkl': 'pickle', 'yml': 'yaml'}
    _types = {'json', 'pickle', 'toml', 'yaml'}

    def __init__(self, document: Document) -> None:
        self._document = document

    @classmethod
    def autoFromBytes(cls, content: bytes, all: bool = False) -> 'Self':
        for type in cls._types:
            try:
                return cls.fromBytes(content, type, all)
            except Exception:
                pass
        raise Exception('Unable to recognize content type')

    @classmethod
    def autoFromString(cls, text: str, all: bool = False) -> 'Self':
        return cls.autoFromBytes(text.encode(), all)

    @classmethod
    def autoFromPath(cls, path: Path, all: bool = False) -> 'Self':
        return cls.autoFromBytes(p.Path(path).read_bytes(), all)

    @classmethod
    def fromDocument(cls, document: Document) -> 'Self':
        return cls(document)

    @classmethod
    def fromBytes(cls, content: bytes, type_or_suffix: str = 'json', all: bool = False) -> 'Self':
        type = cls.typeFromSuffix(type_or_suffix)
        if type not in cls._types:
            raise Exception(f'"{type}" is not a valid type string')
        else:
            return {
                'json': lambda content: cls.fromJSON(content.decode()),
                'pickle': lambda content: cls.fromPickle(content),
                'toml': lambda content: cls.fromTOML(content.decode()),
                'yaml': lambda content: cls.fromYAML(content.decode(), all),
            }[type](content)

    @classmethod
    def fromString(cls, text: str, type: str = 'json', all: bool = False) -> 'Self':
        return cls.fromBytes(text.encode(), type, all)

    @classmethod
    def fromPath(cls, path: Path, all: bool = False, type: t.Optional[str] = None) -> 'Self':
        path = p.Path(path)
        type_or_suffix = path.suffix if type is None else type  # type or path.suffix
        return cls.fromBytes(path.read_bytes(), type_or_suffix, all)

    @classmethod
    def fromJSON(cls, text: str) -> 'Self':
        return cls(json.loads(text))

    @classmethod
    def fromPickle(cls, text: bytes) -> 'Self':
        return cls(pickle.loads(text))

    @classmethod
    def fromTOML(cls, text: str) -> 'Self':
        return cls(tomlkit.loads(text))

    @classmethod
    def fromYAML(cls, text: str, all: bool = False) -> 'Self':
        document = list(yaml.load_all(text)) if all else yaml.load(text)
        return cls(document)

    @classmethod
    def typeFromSuffix(self, type_or_suffix: str) -> str:
        type = type_or_suffix.lstrip('.')
        return self._alias.get(type, type)  # assert _ in self._types

    @classmethod
    def suffixes(cls, dot: bool = True) -> t.Set[str]:
        ans = cls._types | cls._alias.keys()
        if dot:
            ans = set(map(lambda x: f'.{x}', ans))
        return ans

    def to_document(self) -> Document:
        return self._document

    def to_bytes(self, type_or_suffix: str = 'json', all: bool = False, **kwargs: t.Any) -> bytes:
        type = self.typeFromSuffix(type_or_suffix)
        if type not in self._types:
            raise Exception(f'"{type}" is not a valid type string')
        else:
            return {
                'json': lambda: self.to_json(**kwargs).encode(),
                'pickle': lambda: self.to_pickle(**kwargs),
                'toml': lambda: self.to_toml(**kwargs).encode(),
                'yaml': lambda: self.to_yaml(all, **kwargs).encode(),
            }[type]()

    def to_string(self, type: str = 'json', all: bool = False, **kwargs: t.Any) -> str:
        return self.to_bytes(type, all, **kwargs).decode()

    def to_path(self, path: Path, all: bool = False, type: t.Optional[str] = None, **kwargs: t.Any) -> p.Path:
        path = p.Path(path)
        type_or_suffix = path.suffix if type is None else type  # type or path.suffix
        path.write_bytes(self.to_bytes(type_or_suffix, all, **kwargs))
        return path

    def to_json(self, **kwargs: t.Any) -> str:
        kwargs = {'ensure_ascii': False, **kwargs}
        return json.dumps(self._document, **kwargs)

    def to_pickle(self, **kwargs: t.Any) -> bytes:
        return pickle.dumps(self._document, **kwargs)

    def to_toml(self, **kwargs: t.Any) -> str:
        return tomlkit.dumps(self._document, **kwargs)

    def to_yaml(self, all: bool = False, **kwargs: t.Any) -> str:
        kwargs = {'indent': 4, **kwargs}
        return (yaml.dump_all if all else yaml.dump)(self._document, **kwargs)

    auto_from_bytes = deprecated_classmethod(autoFromBytes)
    auto_from_string = deprecated_classmethod(autoFromString)
    auto_from_path = deprecated_classmethod(autoFromPath)
    from_document = deprecated_classmethod(fromDocument)
    from_bytes = deprecated_classmethod(fromBytes)
    from_string = deprecated_classmethod(fromString)
    from_path = deprecated_classmethod(fromPath)
    from_json = deprecated_classmethod(fromJSON, 'from_json')
    from_pickle = deprecated_classmethod(fromPickle)
    from_toml = deprecated_classmethod(fromTOML, 'from_toml')
    from_yaml = deprecated_classmethod(fromYAML, 'from_yaml')
    type_from_suffix = deprecated_classmethod(typeFromSuffix)



__all__ = ['Url']


import pathlib as p
import typing as t
import urllib.parse
import urllib.request

from ..base.type import Dict, Keys
from ..util.decorator import Match
from ..util.function import deprecated_classmethod
from ..util.object.conversion import Conversion

if t.TYPE_CHECKING:
    from typing_extensions import Self

    from ..base.core import Foam


class Url:
    '''Merge remote multiple files into a single file

    Example:
        >>> foam = Foam.fromDemo('cavity')
        Foam.fromPath('.../of.yaml/foam/static/demo/7/cavity.yaml', warn=False)
        >>> foam.save('case')

        >>> data = {
        ...     'name': 'constant/polyMesh',
        ...     'type': ['path', '7z'],
        ...     'permission': None,
        ...     'data': 'static/airFoil2D-polyMesh.7z',
        ... }
        >>> url = Url.fromFoam(foam)
        >>> url.set_url('https://raw.githubusercontent.com/iydon/of.yaml-tutorial/main/tutorials/7/incompressible/simpleFoam/airFoil2D.yaml')
        >>> url[tuple(data['type'])](data)
        {'name': 'constant/polyMesh',
         'type': ['embed', '7z'],
         'permission': None,
         'data': ...}
    '''

    match = Match.default()

    def __init__(self, foam: 'Foam') -> None:
        self._foam = foam
        self._path = None
        self._url = None

    def __getitem__(self, keys: Keys[str]) -> t.Callable:
        if not isinstance(keys, tuple):
            return self.__getitem__((keys, ))
        method = self.match.get(*keys, default=lambda self, static: static)
        return lambda *args, **kwargs: method(self, *args, **kwargs)

    @classmethod
    def fromFoam(cls, foam: 'Foam') -> 'Self':
        return cls(foam)

    @property
    def path(self) -> p.Path:
        return self._path

    @property
    def root(self) -> p.Path:
        return self._path.parent

    @property
    def url(self) -> str:
        return urllib.parse.urlunsplit(self._url)

    def set_url(self, url: str) -> 'Self':
        self._url = urllib.parse.urlsplit(url)
        self._path = p.Path(self._url.path)
        return self

    def set_split_url(self, split_url: urllib.parse.SplitResult) -> 'Self':
        self._url = split_url
        self._path = p.Path(split_url.path)
        return self

    def url_from_path(self, path: p.Path) -> str:
        parts = list(self._url)
        parts[2] = path.as_posix()
        return urllib.parse.urlunsplit(parts)

    @match.register('path', 'raw')
    def _(self, static: Dict) -> Dict:
        # TODO: directory
        url = self.url_from_path(self.root/static['data'])
        static.update({'type': ['embed', 'binary'], 'data': self._urlopen(url)})
        return static

    @match.register('path', '7z')
    def _(self, static: Dict) -> Dict:
        url = self.url_from_path(self.root/static['data'])
        static.update({'type': ['embed', '7z'], 'data': self._urlopen(url)})
        return static

    @match.register('path', 'foam', 'json')
    def _(self, static: Dict) -> Dict:
        return self._path_foam(static)

    @match.register('path', 'foam', 'toml')
    def _(self, static: Dict) -> Dict:
        return self._path_foam(static)

    @match.register('path', 'foam', 'yaml')
    def _(self, static: Dict) -> Dict:
        return self._path_foam(static)

    def _urlopen(self, url: str) -> bytes:
        with urllib.request.urlopen(url) as f:
            return f.read()

    def _path_foam(self, static: Dict) -> Dict:
        url = self.url_from_path(self.root/static['data'])
        self._foam['foam'][static['name'].split('/')] = Conversion \
            .fromBytes(self._urlopen(url), static['type'][2], all=False) \
            .to_document()
        static.update({'type': []})
        return static

    from_foam = deprecated_classmethod(fromFoam)



__all__ = ['Timer', 'TimerResult']


import contextlib
import time
import typing as t
import warnings as w

from ...base.type import Keys
from ...compat.functools import cached_property

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Timer:
    '''Timer

    Example:
        >>> timer = Timer.default()
        >>> with timer.tic_toc('1', '2', '3') as t:
        ...     time.sleep(9)
        >>> print(float(t), timer['1', '2', '3'])
        9.00686868999037 9.00686868999037

    Reference:
        - https://docs.python.org/3/library/time.html
        - https://stackoverflow.com/questions/5849800/what-is-the-python-equivalent-of-matlabs-tic-and-toc-functions
    '''

    def __init__(self, func: t.Callable) -> None:
        self._func = func
        self._cache = {}

    def __enter__(self) -> 'Self':
        return self

    def __exit__(self, type, value, traceback) -> None:
        pass

    def __getitem__(self, labels: Keys[t.Hashable]) -> float:
        if not isinstance(labels, tuple):
            return self.__getitem__((labels, ))
        return -self._cache[labels]

    def __repr__(self) -> str:
        return f'Timer.{self._func.__name__}()'

    def __str__(self) -> str:
        return f'<Timer @ time.{self._func.__name__}>'

    @classmethod
    def best(cls) -> 'Self':
        names = {'monotonic', 'perf_counter', 'process_time', 'thread_time', 'time'}
        order = lambda info: (info.resolution, not info.monotonic, info.adjustable)
        key = lambda name: order(time.get_clock_info(name))
        return getattr(cls, min(names, key=key))()

    @classmethod
    def default(cls) -> 'Self':
        return cls.perf_counter()

    @classmethod
    def monotonic(cls) -> 'Self':
        '''Return the value (in fractional seconds) of a monotonic clock, i.e. a clock that cannot go backwards. The clock is not affected by system clock updates. The reference point of the returned value is undefined, so that only the difference between the results of two calls is valid.'''
        return cls(time.monotonic)

    @classmethod
    def perfCounter(cls) -> 'Self':
        '''Return the value (in fractional seconds) of a performance counter, i.e. a clock with the highest available resolution to measure a short duration. It does include time elapsed during sleep and is system-wide. The reference point of the returned value is undefined, so that only the difference between the results of two calls is valid.'''
        return cls(time.perf_counter)

    @classmethod
    def processTime(cls) -> 'Self':
        '''Return the value (in fractional seconds) of the sum of the system and user CPU time of the current process. It does not include time elapsed during sleep. It is process-wide by definition. The reference point of the returned value is undefined, so that only the difference between the results of two calls is valid.'''
        return cls(time.process_time)

    @classmethod
    def threadTime(cls) -> 'Self':
        '''Return the value (in fractional seconds) of the sum of the system and user CPU time of the current thread. It does not include time elapsed during sleep. It is thread-specific by definition. The reference point of the returned value is undefined, so that only the difference between the results of two calls in the same thread is valid.'''
        return cls(time.thread_time)

    @classmethod
    def time(cls) -> 'Self':
        '''
        Return the time in seconds since the epoch as a floating point number. The specific date of the epoch and the handling of leap seconds is platform dependent. On Windows and most Unix systems, the epoch is January 1, 1970, 00:00:00 (UTC) and leap seconds are not counted towards the time in seconds since the epoch. This is commonly referred to as Unix time. To find out what the epoch is on a given platform, look at gmtime(0).

        Note that even though the time is always returned as a floating point number, not all systems provide time with a better precision than 1 second. While this function normally returns non-decreasing values, it can return a lower value than a previous call if the system clock has been set back between the two calls.

        The number returned by time() may be converted into a more common time format (i.e. year, month, day, hour, etc…) in UTC by passing it to gmtime() function or in local time by passing it to the localtime() function. In both cases a struct_time object is returned, from which the components of the calendar date may be accessed as attributes.
        '''
        return cls(time.time)

    @property
    def cache(self) -> t.Dict[str, float]:
        return self._cache

    @contextlib.contextmanager
    def tic_toc(self, *labels: t.Hashable, builtin: bool = False) -> t.Iterator['TimerResult']:
        if labels in self._cache:
            w.warn(f'Label {labels} already exists')
        try:
            self._cache[labels] = self._func()
            yield self._result(*labels) if builtin else TimerResult(self, *labels)
        finally:
            self._cache[labels] -= self._func()

    def reset(self) -> 'Self':
        self._cache.clear()
        return self

    def wait(self, seconds: float) -> 'Self':
        time.sleep(seconds)
        return self

    @cached_property
    def _result(self) -> type:
        return type('TimerResult', (), {
            '__slots__': ('_labels', ),
            '__init__': lambda this, *labels: setattr(this, '_labels', labels),
            '__float__': lambda this: self.__getitem__(this._labels),
            '__repr__': lambda this: f'TimerResult({self!r}, {", ".join(map(repr, this._labels))})',
            '__str__': lambda this: f'<TimerResult @ ({", ".join(map(repr, this._labels))})>',
            'value': property(lambda this: this.__float__()),
        })

    perf_counter = perfCounter
    process_time = processTime
    thread_time = threadTime


class TimerResult:
    '''Result for timer'''

    __slots__ = ('_timer', '_labels')

    def __init__(self, timer: 'Timer', *labels: str) -> None:
        self._timer = timer
        self._labels = labels

    def __float__(self) -> float:
        return self._timer.__getitem__(self._labels)

    def __repr__(self) -> str:
        return f'TimerResult({self._timer!r}, {", ".join(map(repr, self._labels))})'

    def __str__(self) -> str:
        return f'<TimerResult @ ({", ".join(map(repr, self._labels))})>'

    @property
    def value(self) -> float:
        return self.__float__()



__all__ = ['deprecated_classmethod', 'dict_without_keys', 'dry_run', 'grammar', 'license']


import pathlib as p
import subprocess
import typing as t

from .decorator import message


root = p.Path(__file__).parents[1]


def deprecated_classmethod(cls_meth: classmethod, old: t.Optional[str] = None) -> classmethod:
    func = cls_meth.__func__
    new = func.__name__
    if old is None:
        replace = lambda char: f'_{char.lower()}' if char.isupper() else char
        old = ''.join(map(replace, new)).lstrip('_')
    msg = f'I am using camelCase to distinguish between class methods and instance methods, so please use `{new}` instead of `{old}`'
    return classmethod(message(msg)(func))


def dict_without_keys(data: dict, *keys: str) -> dict:
    # dict := t.Dict[t.Hashable, t.Any]
    return {key: data[key] for key in data.keys() ^ keys}


def dry_run(recover: bool = False) -> None:
    from ..util.object.popen import Origin, DryRun

    subprocess.Popen = Origin if recover else DryRun


def grammar() -> str:
    return (root/'static'/'grammar'/'openfoam.lark').read_text()


def license(full_text: bool = False) -> str:
    if full_text:
        return (root/'static'/'LICENSE.txt').read_text()
    else:
        return 'GPL-3.0-only'



__all__ = ['Data']


import pathlib as p
import typing as t

from .conversion import Conversion
from ..function import deprecated_classmethod
from ...base.type import Dict, FoamItem, Keys, List, Path

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Data:
    '''Multi-key dictionary or list (not recommended)

    Example:
        >>> data = Data.fromDictKeys(
        ...     ('left', 'x'), ('left', 'y'),
        ...     ('right', 'x'), ('right', 'y'),
        ...     default=list,
        ... )
        >>> data['left', 'x'].append({...})
        >>> data['right', 'y'].append({...})

        >>> for key, val in data.items():
        ...     print(key, val)
        ('left', 'x') [{Ellipsis}]
        ('left', 'y') []
        ('right', 'x') []
        ('right', 'y') [{Ellipsis}]
    '''

    def __init__(self, data: FoamItem) -> None:
        self._data = data

    def __contains__(self, keys: Keys[t.Any]) -> bool:
        if isinstance(keys, tuple):
            ans = self._data
            for key in keys:
                if key not in ans:
                    return False
                else:
                    ans = ans[key]
            return True
        elif isinstance(keys, list):
            return self.__contains__(tuple(keys))
        else:
            return self._data.__contains__(keys)

    def __getitem__(self, keys: Keys[t.Any]) -> t.Any:
        if isinstance(keys, tuple):
            ans = self._data
            for key in keys:
                ans = ans[key]
            return ans
        elif isinstance(keys, list):
            # TODO: throw DeprecationWarning
            return self.__getitem__(tuple(keys))
        else:
            return self._data[keys]

    def __setitem__(self, keys: Keys[t.Any], value: t.Any) -> None:
        if isinstance(keys, tuple):
            assert keys

            ans = self._data
            for key in keys[:-1]:
                if isinstance(ans, dict):
                    ans = ans.setdefault(key, {})
                elif isinstance(ans, list):
                    ans = ans[key]
                else:
                    raise Exception(f'Unknown type "{type(ans).__name__}"')
            ans[keys[-1]] = value
        elif isinstance(keys, list):
            # TODO: throw DeprecationWarning
            self.__setitem__(tuple(keys), value)
        else:
            self._data[keys] = value

    def __bool__(self) -> bool:
        return bool(self._data)  # 'list' object has no attribute '__bool__'

    def __iter__(self) -> t.Iterator[t.Any]:
        return self._data.__iter__()

    def __len__(self) -> int:
        return self._data.__len__()

    def __repr__(self) -> str:
        return f'Data({self._data!r})'

    def __str__(self) -> str:
        return self._data.__str__()

    @classmethod
    def fromAny(cls, data: FoamItem) -> 'Self':
        return cls(data)

    @classmethod
    def fromDict(cls, data: t.Optional[Dict] = None) -> 'Self':
        return cls({} if data is None else data)

    @classmethod
    def fromDictKeys(cls, *keys: t.Hashable, default: t.Callable = dict) -> 'Self':
        self = cls.fromDict()
        for key in keys:
            self.__setitem__(key, default())
        return self

    @classmethod
    def fromList(cls, data: t.Optional[List] = None) -> 'Self':
        return cls([] if data is None else data)

    @classmethod
    def fromListLength(cls, length: int, default: t.Callable = lambda: None) -> 'Self':
        return cls.fromList([default() for _ in range(length)])

    @classmethod
    def load(cls, *paths: Path, type: t.Optional[str] = None) -> t.Iterator['Self']:
        for path in map(p.Path, paths):
            type_or_suffix = path.suffix if type is None else type  # type or path.suffix
            yield cls.loads(path.read_bytes(), type_or_suffix)

    @classmethod
    def loadFromPath(cls, *parts: str, type: t.Optional[str] = None) -> 'Self':
        return next(cls.load(p.Path(*parts), type))

    @classmethod
    def loads(cls, content: bytes, type_or_suffix: str = 'yaml') -> 'Self':
        return cls.fromAny(Conversion.fromBytes(content, type_or_suffix, all=True).to_document())

    @property
    def data(self) -> FoamItem:
        return self._data

    def dump(self, *paths: Path, type: t.Optional[str] = None) -> 'Self':
        for path in map(p.Path, paths):
            type_or_suffix = path.suffix if type is None else type  # type or path.suffix
            path.write_bytes(self.dumps(type_or_suffix))
        return self

    def dump_to_path(self, *parts: str, type: t.Optional[str] = None) -> 'Self':
        return self.dump(p.Path(*parts), type)

    def dumps(self, type_or_suffix: str = 'yaml', **kwargs: t.Any) -> bytes:
        return Conversion \
            .fromDocument(self._data) \
            .to_bytes(type_or_suffix, all=True, **kwargs)

    def is_dict(self) -> bool:
        return isinstance(self._data, dict)

    def is_list(self) -> bool:
        return isinstance(self._data, list)

    def is_other(self) -> bool:
        return not (self.is_dict() or self.is_list())

    def to_any(self) -> FoamItem:
        return self._data

    def to_dict(self) -> Dict:
        assert self.is_dict()

        return self._data

    def to_list(self) -> List:
        assert self.is_list()

        return self._data

    def contains(self, *keys: t.Any) -> bool:
        try:
            return self.__contains__(keys)
        except Exception:
            return False

    def get(self, key: t.Any, default: t.Optional[t.Any] = None) -> t.Any:
        # TODO: retained due to compatibility needs
        if isinstance(self._data, dict):
            return self._data.get(key, default)
        elif isinstance(self._data, list):
            try:
                return self._data[key]
            except (IndexError, TypeError):
                return default
        else:
            raise Exception(f'Unknown type "{type(self._data).__name__}"')

    def gets(self, *keys: t.Any, default: t.Optional[t.Any] = None) -> t.Any:
        try:
            return self.__getitem__(keys)
        except (IndexError, KeyError, TypeError):
            return default

    def setdefault(self, key: t.Any, default: t.Optional[t.Any] = None) -> t.Any:
        # TODO: retained due to compatibility needs
        return self._data.setdefault(key, default)

    def set_default(self, *keys: t.Any, default: t.Optional[t.Any] = None) -> 'Self':
        if keys not in self:
            self.__setitem__(keys, default)
        return self.__getitem__(keys)

    def set_via_dict(self, data: Dict) -> 'Self':
        for keys, value in self.__class__.fromDict(data).items():
            self.__setitem__(keys, value)
        return self

    def items(self, with_list: bool = False) -> t.Iterator[t.Tuple[Keys[t.Any], t.Any]]:
        yield from self._items(self._data, with_list=with_list)

    def _items(
        self,
        data: t.Any, with_list: bool = False, keys: Keys[t.Any] = (),
    ) -> t.Iterator[t.Tuple[Keys[t.Any], t.Any]]:
        if isinstance(data, dict):
            for key, value in data.items():
                yield from self._items(value, with_list, keys+(key, ))
        elif with_list and isinstance(data, list):
            for key, value in enumerate(data):
                yield from self._items(value, with_list, keys+(key, ))
        else:
            yield keys, data

    from_any = deprecated_classmethod(fromAny)
    from_dict = deprecated_classmethod(fromDict)
    from_dict_keys = deprecated_classmethod(fromDictKeys)
    from_list = deprecated_classmethod(fromList)
    from_list_length = deprecated_classmethod(fromListLength)
    load_from_path = deprecated_classmethod(loadFromPath)



__all__ = ['Version']


import typing as t

from ..function import deprecated_classmethod

if t.TYPE_CHECKING:
    from typing_extensions import Self


class Version(t.NamedTuple):
    '''Version named tuple

    Example:
        >>> version = Version.fromString('1.2.x')
        >>> version.major, version.minor, version.other, version.micro
        (1, 2, 'x', None)
        >>> version.to_string()
        '1.2.x'
    '''

    major: int
    minor: int
    other: t.Optional[str]

    def __lt__(self, other: 'Self') -> bool:
        return (self.major, self.minor) < (other.major, other.minor)

    def __gt__(self, other: 'Self') -> bool:
        return other.__lt__(self)

    def __repr__(self) -> str:
        return f'Version.fromString({self.to_string()!r})'

    def __str__(self) -> str:
        return self.to_string()

    @classmethod
    def fromString(cls, version: str) -> 'Self':
        parts = version.split('.', maxsplit=2)
        if len(parts) == 2:
            major, minor = parts
            other = None
        elif len(parts) == 3:
            major, minor, other = parts
        else:
            raise Exception(f'"{version}" is not a valid version string')
        return cls(int(major), int(minor), other)

    @property
    def micro(self) -> t.Optional[int]:
        if self.other is not None:
            parts = self.other.split('.')
            if parts and parts[0].isdigit():
                return int(parts[0])
        return None

    @property
    def micro_int(self) -> int:
        return self.micro or 0

    def to_string(self) -> str:
        version = f'{self.major}.{self.minor}'
        if self.other is not None:
            version += f'.{self.other}'
        return version

    from_string = deprecated_classmethod(fromString)
